
/* Portions Copyright (c) 1995,1996,1997 by Trusted Information Systems, Inc.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND TRUSTED INFORMATION SYSTEMS DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL TRUSTED INFORMATION 
 * SYSTEMS BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 *
 * Trusted Information Systems, Inc. has received approval from the
 * United States Government for export and reexport of TIS/DNSSEC
 * software from the United States of America under the provisions of
 * the Export Administration Regulations (EAR) General Software Note
 * (GSN) license exception for mass market software.  Under the
 * provisions of this license, this software may be exported or
 * reexported to all destinations except for the embargoed countries of
 * Cuba, Iran, Iraq, Libya, North Korea, Sudan and Syria.  Any export
 * or reexport of TIS/DNSSEC software to the embargoed countries
 * requires additional, specific licensing approval from the United
 * States Government. 
 */




/* 
 * $Header: /home/ogud/sec_bind/signer/RCS/signer.c,v 1.36 1997/01/22 22:06:35 ogud Exp $
 */
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <fcntl.h>
#include <sys/stat.h>
#define DEBUGFILE "Signer.out"
#define MAIN_PROGRAM
#include "named.h"
#include "dnssec.h"
#undef MAIN_PROGRAM

#define SIGN_TIME (records_signed ?  sum_time / (double) records_signed : (double) 0.0)
/*
 * Local Function definitions 
 */
static void     do_output        __P((  )),
                output_zone      __P(( char *, char *)),
                set_prefix       __P((int32_t, int32_t, char *, int,int32_t )),
                usage            __P((  ));

static int      make_nullkey     __P(( char * , struct databuf *, char *)),
                save_name        __P(( FILE *, char *, int, int , char *)),
                sign_name        __P(( nxt *, u_int32_t, char *, int));

static u_int32_t set_seq_num      __P(( time_t));

static char     *sign_init        __P(( int, char **));

static int           self_sign_pubkey=0;
static char          *boot_file, *out_dir; 

char                 *chache_file; 
static int            verify_flag = 0;
static unsigned int   expr_val    = 3600*24*7;  /* only used in Signer.c */
static KEY           *SIGN_KEY=NULL; /* only used in Signer.c */
static struct timeval last_time, time_now;
static int            records_signed = 0;
static int            report_success = 1;
static int            gen_nullkey = 0;
static double         sum_time = 0.0;
static int            do_nxt=1;       /* make NXT records */
static char           *key_list[256], *zone[256];
static int            max_zones=0;

int 
main(argc, argv)
int     argc;
char  **argv;
{
/* initalize */
    default_signer = sign_init(argc, argv);
    (void) printf("\nSigner: %s expr=%s \n", default_signer, 
		  ctime((time_t *) &expr_val));
    
    if( !do_not_sign && !dnssec_check(KEY_RSA)) { /* no crypto */
      printf("\nSigner: Signing disabled crypto unavailable for alg %d\n",
	     KEY_RSA);
      do_not_sign = 1; 
    }
      
    printf("Signer:Invoking boot_read\n");
    (void) ns_init(boot_file);
              /* Now I can start outputing records */
    printf("Signer:Done signing\n\n");
              /* output handling */ 
    do_output();
    dumpfile = "signer_out";
    doadump();
    return 0;
  }

/* 
 * save_name(): This function writes out the requested RR's for a given name
 *  request can be for : 
 *         soa  : Prints out SOA, SIG(SOA) and SIG(AXFR)
 *        not soa: Prints out all RR's but the ones soa request prints out
 *  We can also have this routine print out the delegation RR for a subzone. 
 *     when writing out normal RR's: static variables {ns,key,nxt}_found 
 *     are set. When delegation flag is set an attempt is made to print 
 *     these RR's.
 *  Returns 0 if this name is not a delegation point
 *          1 if this IS a delegation point
 *
 */
static int origin_len=0;
static struct databuf *ns_found=NULL, *key_found=NULL, *nxt_found=NULL; 
static int 
save_name( fp, name, soa, delegation, out_name)
     char *name, *out_name;
     FILE * fp;
     int soa,delegation;
{
    u_int16_t t;
    struct namebuf *np;
    struct databuf *dp=NULL, *adp=NULL;        /* data records */
    char *sep="\t;";
    dp = adp = NULL;
    if( soa) {
      dp = nt_lookup( name, T_SOA, &np);
      if( do_axfr) {  /* find the SIG(AXFR) RRR if there is one */
	  for( adp = dp ? dp : nt_lookup(name, 0, &np);  adp; 
	      adp = adp->d_next) 
	    if( adp->d_type == T_SIG && SIG_COVERS(adp) == T_AXFR)
	      break;
	} /* if(do_axfr)   */
    }
    else if(!delegation)   
      dp = nt_lookup( name, 0, &np);
/* now lets do some error checking */
    if( delegation) {
      if( !ns_found || !key_found) {
        printf("save_name:Missing NS %hx or KEY %hx RR at delegation %s\n",
               ns_found, key_found, name);
        ns_found = key_found = nxt_found = NULL;
        return 0;
      }
    }
    else if( np == NULL) {
      (void) printf("save_name:NO such name %s\n", name); 
      return 0; 
    }
    else if( dp == NULL) {
      if(soa) (void) printf("save_name:NO SOA for +%s+\n", name);
      else    (void) printf("save_name:NO RR  for +%s+\n", name);
      if( !adp) return 0; 
    }
    if( key_file) { 
      if( soa) {
	fprintf(fp,"\n%s",out_name);
	if( strchr(out_name,'.')) putc('.', fp);
      }
      else if( *out_name != '@') {
	(void) fprintf(fp, "\n%s", out_name);
	if( strchr(out_name,'.')) putc('.', fp);
      }
    }
    else if( soa || *out_name != '@') fprintf(fp, "\n%s",out_name);
     /* if name out of $orgin range terminate it with a dot */

    if(delegation) {  /* this writes out the PARENT file */
      dp = nt_lookup( name, 0, &np);
      if( key_found) {  /* I can not count on first key to be key_found */
	for( dp ; dp && dp->d_type; dp = dp->d_next) 
	  if(dp->d_type == T_KEY) {
	    fprintf( fp, "\t%d ", dp->d_ttl);
	    print_rr(fp, dp, sep);         /* writing out KEY RR's */
	  }
	  else if ((dp->d_type == T_SIG) && (SIG_COVERS(dp) == T_KEY)) 
	    key_found = dp;                /* found SIG(KEY) write out later */
	if( key_found) {
	  fprintf( fp, "\t%d ", key_found->d_ttl);
	  print_rr(fp, key_found, sep);      /* write out SIG(KEY) */
	}
      }
      if( nxt_found) {
	for( dp = np->n_data; dp ; dp = dp->d_next) 
	  if( dp->d_type == T_NXT ){
	    fprintf( fp, "\t%d ", dp->d_ttl);
	    print_rr(fp, dp, sep);          /* writing out NXT */
	  }
	  else if((dp->d_type == T_SIG) && (SIG_COVERS(dp) == T_NXT))
	    nxt_found = dp;
	fprintf( fp, "\t%d ", nxt_found->d_ttl);
	print_rr( fp, nxt_found, sep);
      }
      ns_found = key_found = nxt_found = NULL;
    }
    else {
      ns_found = key_found = nxt_found = NULL;
      if( soa) { /* find the SOA SIG(SOA) and SIG(AXFR) for this name 
		  to put at top of a file*/
	if( dp) {
	  print_rr(fp, dp, sep);  /* writing out SOA */
	  dp = dp->d_next;
	  if( dp && dp->d_type == T_SIG) 
	    print_rr(fp, dp, sep);     /* writing out SIG(SOA) */
	}
	if( adp) 
	  print_rr(fp, adp, sep);
	fputc('\n',fp);
      }   /* if soa && dp */
      else { 
	for( ; dp != NULL; dp = dp->d_next) {
	  if( dp->d_type == T_SOA) { /* SKIP SOA and SIG(SOA) */
	    if( dp->d_next  && (dp->d_next->d_type == T_SIG) )
	      dp = dp->d_next; 
	    continue;
	  }
	  else if( (dp->d_type == T_SIG) && (t = SIG_COVERS(dp)) == T_AXFR) 
	    continue; /* skip SIG(AXFR) */
	  print_rr(fp, dp, sep);
	  if( dp->d_type == T_NS) ns_found=dp;
	  else if( dp->d_type == T_KEY && ns_found) {
	    /* can only set key flag if we have found ns RR */
	    u_int16_t key = _getshort(dp->d_data);
	    if( key & (SIGN_F_SIGNATORY | SIGN_F_ZONE))
	      key_found=dp;
	  }
	  else if( dp->d_type == T_NXT && key_found)
	    nxt_found=dp;
	}            /* for */
      }
    }
    (void) fputc('\n',fp);
    if( key_found)  /* implies ns was found */
      return 1;
    return 0;
}

/* 
 * do_output() Creates the correct output files user supplies an output 
 * directory or default /tmp, this routine copies the conf.file there,
 *  then we write out the zone records out to there in correct files
 */

static void 
do_output()
{
  char *name, out_name[PATH_MAX];
  char *t, *n, keyname[PATH_MAX], zone_name[PATH_MAX];
  char tmp[1024];
  char long_string[500];
  FILE  *inf, *outf, *ikf;
  extern int errno;
  int i=0;

  if( out_dir == NULL) 
    return  ;
  name = (char *) strrchr( boot_file, '/'); 
  if( name == NULL) name = boot_file;
  else name++;
  if((inf = fopen(name,"r")) ==NULL){
    printf("Signer:do_output(): %sFailed opening file=%s code=%d\n", 
	   boot_file, name, errno); 
    return;
  }
  sprintf(out_name,"%s/%s", out_dir, name);
  if( access(out_name,F_OK) ==0 )
    unlink(out_name);
  if( ( outf = fopen( out_name, "w+")) == NULL) {
    printf("Signer: Failed creating new boot file=%s= %d\n", 
	   out_name, errno); 
    perror("File creation error");
    return;
  }
  /* copy boot file to output directory */
  while( fgets(tmp,sizeof(tmp),inf)) {
    if( (strncasecmp(tmp,"primary",7) ==0) ||
        (strncasecmp(tmp,"keyfile",7) ==0)) {
      memset( zone_name, 0, sizeof(zone_name));
      for(t=tmp+7;isspace(*t); t++);  /* skip spaces */
      for(n=zone_name; isalnum(*t) ||*t =='.' || *t == '-' || *t =='_';) 
	*n++=*t++; /* copy name*/
      if( zone[i] && (strcasecmp(zone[i],zone_name) ==0)) {
	if( key_list[i]) {
	  int j;
	  for(j=0; j < i && 
	      (key_list[j]== NULL || strcasecmp(key_list[i],key_list[j])); 
	       j++); 
	  if( j == i) {  /* this  is not a duplicate */
	    sprintf(keyname, "%s.public",key_list[i]);
	    if( (ikf =fopen(keyname,"r"))== NULL)
	      printf("Signer:do_output():file=%s:Failed opening %s code=%d\n", 
		     boot_file, keyname, ikf); 
	    else {
	      memset(long_string,0, sizeof(long_string));
	      fgets(long_string,sizeof(long_string),ikf);
	      fclose(ikf);
	      fprintf(outf,"pubkey %s. %s;\n", key_list[i],long_string);
	    }   /* else */
	  }     /* if ( j ==i */
	}       /* if key_list[] */
	i++;
      }         /* if( strcase */
    }           /* if( primary */
    else if( strncasecmp(tmp, "directory", 9)==0) { /* change dir */
      for( t=&tmp[9];isspace(*t);t++);
      if( out_dir[0] == '/')
	sprintf(t,"%s\n", out_dir);
      else if( *t == '/'){
	while( !isspace(*t))  t++;
	sprintf(t,"/%s\n", out_dir);
      }
      else { 
	char *p = (char *) getcwd(NULL, sizeof(tmp));
	sprintf( t,"%s/%s\n",p,out_dir);
	free(p);
      }
    }
    else if( strncasecmp(tmp,"pubkey",6)==0) 
      continue;
    fputs(tmp, outf);
  }          /* while ( fgets */
  fclose(outf);
  fclose(inf);
  return;
}      /* do_output */


/* 
 * output_zone(): This function writes out all the RR's in a zone
 *  first it creates appropriate file
 * When we detect a name in the zone that is a delegation point we write
 * the delegation RR's to a file named <name>.<zonename>.PARENT
 * At the end of each zone file we add and $INCLUDE <zonename>.PARENT to
 * Include delegation RR's signed by our parent, otherwise we the
 * parent delegation RR's will get lost when someone loads this zone
 * For correct opperation the delegation RR's (NS and KEY) better be the 
 * same or totaly missing in this zone.  
 */ 

static void
output_zone( filename, zonename)
char *filename, *zonename;
{
  nxt             *s;
  struct databuf *dp;
  FILE           *fp, *fp_p;
  char           full_name[PATH_MAX], *name;
  int            len; 

  if( filename == NULL || zonename == NULL) {
    printf("Signer.c:output_zone(): Bad input File=+%s+ Zone=+%s+\n",
	   filename, zonename);
	   return;
  }
  if( nxt_start == NULL) {
    printf("Signer.c:output_zone(): No list to output boot file %s\n",
	   boot_file);
    return ;
  }
  len = strlen(out_dir); 
  if( len+strlen(filename)+1 > sizeof(full_name)) {
    printf("output_zone:Output path to long %s/%s\n",out_dir,filename);
    exit(-99);
  }
  sprintf(full_name,"%s/%s",out_dir, filename);
  printf("output_zone: Writing file +%s+\n", full_name);
  if( (fp = fopen( full_name, "w")) == NULL) {
      perror("Failed opening output file");
      printf("Filename: %s\n", full_name);
      printf("Current directory Filename: %s\n", getcwd(NULL,200));
      exit( 23);
  }
  if(strlen(zonename))
    fprintf(fp, "$ORIGIN %s.\n",zonename);
  else 
    fprintf(fp, "$ORIGIN .\n");
  origin_len = strlen(zonename);
  if(origin_len == 0) origin_len--;
  (void) save_name(fp, zonename, DO_SOA, 0, "@") ;
  for(dp =key_found; dp && dp->d_type == T_KEY; dp=dp->d_next);
  if( dp && dp->d_type == T_SIG &&  SIG_COVERS(dp) == T_KEY) {
    char *fmt="; ****WARNING Self signed pubkey=%s SIG(AXFR's) MAY fail\n";
    fprintf(fp,  fmt, zonename);
    fprintf(stderr, fmt, zonename);
  }
  (void) save_name(fp, zonename, NO_SOA, 0, "@") ;
  if( key_file && strcasecmp(zonename,nxt_start->dname))  s = nxt_start;
  else s = nxt_start->next;
  for( ; s != NULL; s = s->next) {
    if( !key_file) name = s->dname;
    else {
      int nn = strlen(s->name);
      if( nn >= origin_len && 
	strncmp( &s->name[nn-origin_len], zonename, origin_len) ==0)
	name = s->dname;
      else name = s->name;
    }
    if(save_name( fp, s->name, NO_SOA, 0, name) && 
         SIGN_BIT_ISSET( T_KEY, s->nxt_rr)) {
      /* delgation point with signed KEY record */
      sprintf(full_name,"%s/%s.PARENT",out_dir, s->name);
      printf("output_zone: Writing file +%s+\n", full_name);
      if( (fp_p = fopen(full_name, "w")) == NULL) {
	perror("Failed opening output file");
	printf("Filename: %s\n", filename);
	exit( 24);
      }
      fprintf(fp_p,"$ORIGIN %s.\n",zonename);
      (void) save_name(fp_p, s->name, NO_SOA, 1, name);
      fclose(fp_p);
    }
  }
  if(!key_file) 
    fprintf(fp,"$INCLUDE %s.PARENT\n",zonename);
  fclose(fp);
}

/*
 * This function will step through a zone and generate the SIG records
 * required for each name.  
 * Inputs:
 *      zp pointer to the zoneinfo structure used to set the flags on
 *         the newly generated NXT and SIG rr's
 *      boot_file: File name zone was loaded from 
 *      origin: name of zone 
 * 
 */
void 
sign_zone(zp, boot_file, origin)
     struct zoneinfo *zp;
     char *boot_file, *origin;
{
  int         clev, axfr, ret,t;
  u_int32_t   ttl = ttl_val
 /*_list[zone_names-1]*/;
  nxt         *s;  /* Pointer to names */
  char        *signer;
  struct namebuf *np;
  struct databuf *dp;        /* data records */

  (void) printf( "\nFinished reading in boot_file %s %d\n", boot_file,ttl);

  clev = dn_count_labels(origin);
  if( ttl == 0) ttl = 10000; /* almost as good */
  
 /*special case to handle root key */ 
  zone[max_zones++] = strlen(origin) ? savestr(origin) : ".";
  if( !nxt_start) { 
    printf( "sign_zone(): NO records match pattern %s\n",origin);
    return ;  /* no records to sign  bail out */
  }
  axfr = do_axfr;
  if( key_file==0 && do_nxt) 
    (void) nxt_make( origin, ttl, zp->z_sec);
  signer = nxt_start->signer;
  if( signer != NULL) {
    if( !key_file && SIGN_BIT_ISSET(T_NS, nxt_start->nxt_rr) &&
       (strncasecmp( origin, signer, strlen(origin)) != 0)){
      printf("Signer:sign_zone: %s\n\torigin= %s\n\tkeyname=%s\n", 
	     "signing zone with key that is not same as origin", 
	     origin, signer);
    }
    if( signer[0] == '.')
      signer++;
    t =strlen(signer);  /* make sure no keyname ends with a . */
    if( signer[--t] == '.') signer[t]='\0';
    key_list[max_zones-1] = savestr(signer);
  }
  if( do_not_sign || signer == NULL) 
      (void) printf( "sign_zone: SIGNING Disabled\n");
  else {  /* I need to sign */
    if( (SIGN_KEY = dnssec_getkey( signer, !DNSSEC_FORCE_READ))) {
      set_prefix(ttl, expr_val, signer, (int) SIGN_KEY->foot, clev);
      init_signer(SIGN_KEY);
      /* Now I can start signing records for each zone */ 
      for( s = nxt_start; s != NULL; s= s->next) {
	if( (ret= sign_name( s, ttl, signer, clev))) 
	  printf( "sign_zone():Error %3d in signing name %s\n", ret,
		 s->dname);
      }            
      if( SIGN_KEY->axfr) { /* now gerate the SIG(AXFR) for the zone */
	dp = nt_lookup(origin, T_SOA, &np);
	if( dp == NULL) 
	  dp = nt_lookup( signer, T_KEY, &np);
	if( dp)
	  do_signature(origin, T_AXFR, dp, signer, SIGN_KEY);
	else {
	  printf("Sign_zone: %s can not find SOA for zone or KEY for signer\n",
		 boot_file);
	  if( key_file) printf(
     "Sign_zone: %s Adding signer KEY to the key file will fix this problem\n"
			       , boot_file);
	}
      }
      printf( "Sign_zone: %s #signatures %d key %d average time %6.3lf\n",
	     boot_file, records_signed, SIGN_KEY->key_size, SIGN_TIME);
    }
    else 
	printf("sign_zone():NO KEY found for signer %s zone %s\n",
	       signer, origin);
  }     /* else {    */
  output_zone(boot_file, origin);
}

static void 
usage() {
  printf("\nSigner: Signer expire [+=]<value bootf <file> \\\n");
  printf("\t\tout-dir <dir> verify debug-file <name>\n");
  printf("Signer: expire [+=]<value>\t;+ time from now is default\n");
  printf("Signer: \t\t\t;= exact time \n");
  printf("Signer: boot-f\t\t<file>\t; name of input file\n");
  printf("Signer: debug-file\t<name>\t; name of file for debug output\n");
  printf("Signer: hide\t\t\t; Hide names in zones with wild card records\n");
  printf("Signer: noaxfr\t\t\t; Do not create AXFR signature for zones\n");
  printf("Signer: nonxt\t\t\t; Do not generate NXT records\n");
  printf("Signer: noreport\t\t;Do not report on each signature generated\n");
  printf("Signer: nosign\t\t\t; Do not sign data\n");
  printf("Signer: nullkey\t\t\t; Create Null keys for delegations without KEY RR\n");
  printf("Signer: out-dir\t<dir>\t\t; where to create the output \n");
  printf("Signer: selfsign\t\t; self sign zone key\n");
  printf("Signer: signer-name\t<name>\t; Default key to use\n");
  printf("Signer: -d <debug level>\t; Sets named debug level\n");
  exit(3);
}

/*
 * sign_init(): function to process arguments, set default values and do 
 *              routine checks of avaliablity of files specified
 */ 
struct timeval  tp;

static char *
sign_init(argc, argv)
	int             argc;
	char          **argv;
{
	u_char         *cp;
	int             len, i;
#ifdef TESTING 
	char *signer =  "foo.bar";
	boot_file = "simple_test/test.boot";      /* default boot file */
	debugfile = "Signer.out";
#else 
	char *signer =  NULL;
	boot_file = NULL;
#endif
        out_dir = "/tmp";
	hide = 0;                  /* wildcards do not hide names by default */
	do_not_sign = 0; 
	gen_nullkey = 0;
	debug = 1; 
	(void) gettime(&tp);     /* get current time in GMT */
	expr_val = tp.tv_sec;              
	/*signer should set the SOA Serial # as a function of time*/
	signer_serial = set_seq_num(tp.tv_sec);

	for(i = 1;  i < argc; i++) {        /* process arguments */
	  if( !strcasecmp(argv[i], "-d")) 
	      if( isdigit(argv[i+1][0]))  debug =atoi(argv[++i]);
	      else debug =1 ;              /* default case if no level */
	  else if( !strcasecmp(argv[i], "boot-f"))       boot_file = argv[++i];
	  else if( !strcasecmp(argv[i], "debug-file"))   debugfile=argv[++i];
	  else if( !strcasecmp(argv[i], "hide"))         hide = 1; 
	  else if( !strcasecmp(argv[i], "noaxfr"))       do_axfr=0;
	  else if( !strcasecmp(argv[i], "nonxt"))        do_nxt = 0; 
	  else if( !strcasecmp(argv[i], "noreport"))     report_success = 0;
	  else if( !strcasecmp(argv[i], "nosign"))       do_not_sign = 1; 
	  else if( !strcasecmp(argv[i], "nullkey"))      gen_nullkey = 1; 
	  else if( !strcasecmp(argv[i], "out-dir"))      out_dir = argv[++i];
	  else if( !strcasecmp(argv[i], "verify"))       verify_flag= 1; 
	  else if( !strcasecmp(argv[i], "seq-no")) {
	    if( isdigit(argv[i+1][0]))  signer_serial = atoi(argv[++i]);
	    else printf("Signer:sign_init(): Bad sequence number %s\n",
			argv[i+1]);
	  }
	  else if( !strcasecmp(argv[i], "selfsign")) {
	    self_sign_pubkey =1;
	    fprintf(stderr, 
		    "sign_init:WARNING pubkeys will be self signed WARNING\n");
	    fprintf(stderr,
     "sign_init:WARNING only do this if your superzone is UNSECURE WARNING\n");
	  }
	  else if( !strcasecmp(argv[i], "expire")){
	    /* I need to check if the first character is =+ and do 
	       correct time managemment */ 
	    cp = (u_char *) argv[++i];
	    if( isdigit(*cp)) expr_val = atoi(cp);
	    else if(*cp=='=') expr_val = atoi(++cp);
	    else if(*cp=='+') expr_val += atoi(++cp);
	    else {
	      printf("%s sign_init():Bad Expiration time %s\n",
		     argv[0], argv[i]);
	      usage();
	    }
	  }
	  else signer = (char *) savestr(argv[i]);
	}  /* for (i= */
/* now I got all the flags make sure the required ones are not missing */ 
	if( signer !=NULL) 
	  printf("%s:Default Signer is %s\n", argv[0], signer);
/*	else 
	  printf("%s: No default Signer specified\n",argv[0]); */
	if(boot_file==NULL) {
	  printf("%s:Boot file required\n",argv[0]);
	  usage();
	} else if( access( boot_file, R_OK|F_OK)){
	  printf("%s:Can not access boot file %s\n",argv[0], boot_file);
	  usage();
	}
	
/* check if exiration time is in the future */
	if( expr_val < tp.tv_sec) {
     printf("sign_init():Can not sign records that expire in the past %d %d\n",
		 expr_val, tp.tv_sec);
	  usage();
	} else if( expr_val == tp.tv_sec) {
	  expr_val += DNSSEC_EXPIRE; /* default records live for 30 days */
	  printf("sign_init(): Records set to expire in %d days\n", 
		 (int) ((expr_val-tp.tv_sec)/3600/24));
	}
	  
/* If name does not start with / I can only perform this check after 
   directory diretive in boot file. */
   
	if( out_dir[0] == '/') {
	  if( access( out_dir, W_OK)) {
	    perror("Signer:output direcory");
	    printf( "%s:No write access to output directory %s\n",
		   signer, out_dir);
	    usage();
	  }
	}
	if( debugfile ) ddt = freopen(debugfile,"w+",stderr);
	else            ddt = freopen("/dev/tty","w+",stderr);
	if( ddt == NULL) {
	  printf("Signer: Can not open /dev/tty as error device");
	  exit (34);
	}
#if defined(HAVE_SETVBUF)
	setvbuf(ddt, NULL, _IOLBF, BUFSIZ);
#else
	setlinebuf(ddt);
#endif
	if ((i = fcntl(fileno(ddt), F_GETFL, 0)) < 0) 
	  printf("Signer:fcntl(ddt, F_GETFL): %d",i);
	else 
	  (void) fcntl(fileno(ddt), F_SETFL, i|O_APPEND);

	/* Get signers private key name */ 
	len = signer ? strlen(signer): 0;
	while( len>0 && signer[len-1] == '.') 
	  signer[--len]='\0';
	return signer; 
} /* sign_init()  */


/* 
 * sign_name(): generates SIG records for one name 
 *              this function uses the sign bit map in the nxt record to
 *              direct it on what RR's to sign. 
 *              This function calculates the AXFR signature in parallel
 *  inputs: nxt_ptr: Pointer to to the nxt record for this name 
 *          ttl:     Used in the signature prefix
 *          zone_signer: name of the default signer unless another one is
 *                   specified in the nxt record;
 *  return values: 
 *      0:  SIG's in the zone database
 *      1:  Nothing to sign
 *     -1:  No signer
 *     -2:  No RR fround for the name in the nxt record
 *     -3:  No key found for the signer specified 
 *     -4:  Key found but not allowed to sign RR's
 *     -5:  No key in database under signer name (duplicate error)
 *     -6:  Failed initalizing Hash function 
 *     -7,-8,-9:  Failed signing data (different locations)
 */
static int
sign_name( nxt_ptr, ttl, zone_signer, clev)
nxt *nxt_ptr;
u_int32_t ttl; 
int clev;
char *zone_signer;
{
    struct hashbuf *htp=hashtab;          /* lookup relative  */
    int type, tput, ret;
    const char *fname;
    struct namebuf *np, *snp;
    struct databuf  *dp;        /* data records */
    static char *current_signer = NULL;
    char *name, *signer;

    if( nxt_ptr->max_type <= 0) 
            return 1; /* nothing to sign */
    signer = nxt_ptr->signer ? nxt_ptr->signer : zone_signer;
    if( !signer ){ 
      printf( "sign_name: NO signer listed for %s\n", nxt_ptr->dname);
      return -1; 
    }
    if( signer[0] == '.') signer++;
    name = nxt_ptr->name;
    snp= np = nlookup(name,  &htp, &fname, 0); /* find records */
    if( np == NULL) {
      printf("sign_name():NO such name %s\n", name); 
      return -2; 
    }
           /* Get the correct key loaded */
    if( !current_signer || strcasecmp( current_signer, signer) ) {
	current_signer = NULL;
      if( (SIGN_KEY = dnssec_getkey( signer, !DNSSEC_FORCE_READ)) == NULL) {
	printf("sign_name():No Key found for signer %s\n",signer);
	return -3; 
      }
      current_signer = savestr( signer);
    }
	/* Generate the signatures */
    if( key_file) {
      for(dp= nt_lookup(name, T_KEY, &np); dp && dp->d_type == T_KEY; 
	  dp= dp->d_next)
	put_rr(name, dp, sig_data.ttl);
      expand_rrs(name); 
      ret = dnssec_sign( SIG_MODE_AXFR, sig_data.data, sig_data.dat_len, 
		      SIGN_KEY, NULL); 
      sig_data.dat_len = 0;
      dp_rrcount = 0;
      return ret;
    }
    set_prefix( ttl, expr_val, signer, (int) SIGN_KEY->foot, dn_count_labels(name));
	
    if( SIGN_BIT_ISSET( T_SOA, nxt_ptr->sign_rr)) {
      if( (dp = nt_lookup(name, T_SOA, &np))) {
	put_rr(name, dp, ttl); 
	if((ret=do_signature(name, T_SOA, snp->n_data, signer, SIGN_KEY))){
	  fprintf(stderr, "sign_name(): %s unable to sign data %d %d\n", 
		  name,type,ret);
	  return ret^0x800;
	}
      }
      SIGN_BIT_CLEAR(T_SOA, nxt_ptr->sign_rr); /* prevent resigning */ 
      if( self_sign_pubkey==0 && SIGN_BIT_ISSET(T_KEY,nxt_ptr->sign_rr))
	SIGN_BIT_CLEAR(T_KEY, nxt_ptr->sign_rr); /* Do not self sign zone key*/
      else if( self_sign_pubkey) {
	/* Need to check if the .PARENT file exists if it is older than 
	   the start time of this signer then can I owerwrite it,
	   or if the if file is not writeable XXXX ogud
	   */
	char full_name[MAXDNAME];
	struct stat sb;
	int j;
	sprintf(full_name, "%s/%s.PARENT", out_dir, name);
	/* two cases  NO PARENT file  */
	j =access(full_name, R_OK);
	if( j ==0 && access(full_name, W_OK)==0) {
	  stat(full_name, &sb);
	  if( sb.st_mtime < tp.tv_sec) j = 1;
	}
	if( j) {
	    SIGN_BIT_SET(T_KEY, nxt_ptr->sign_rr); /* self sign zone key*/
	    SIGN_BIT_SET(T_KEY, nxt_ptr->nxt_rr); /* self sign zone key*/
	  }
      }
    }
    /* do not sign NS  unless these are your NS RR's */
    else {
      u_int32_t sign=0;
      SIGN_BIT_CLEAR(T_NS, nxt_ptr->sign_rr);
      if( SIGN_BIT_ISSET( T_KEY, nxt_ptr->nxt_rr)) /* Check KEY's */
	for( dp= nt_lookup( name, T_KEY, &np) ; dp ; dp = dp->d_next)
	  if( dp->d_type == T_KEY) {
	    u_int32_t flags = (u_int32_t) _getshort(dp->d_data);
	    sign |= (flags &( SIGN_F_ZONE | SIGN_F_SIGNATORY));
	  }
      if ((sign != 0) && !SIGN_BIT_ISSET( T_NS, nxt_ptr->nxt_rr))
	printf( "sign_name: %s has signing authority but NO NS record\n",name);
      else if( gen_nullkey && sign == 0 && 
	      SIGN_BIT_ISSET( T_NS, nxt_ptr->nxt_rr))
	/* NS present but NO key */
	make_nullkey(name, np->n_data, signer); 
    }

    /* sign all remaining RR's  */
    for( tput=0, dp = snp->n_data; dp ; dp = dp->d_next) { 
      if( SIGN_BIT_ISSET(dp->d_type, nxt_ptr->sign_rr)) {
	if( type != dp->d_type && tput>0) {
	  if((ret=do_signature(name, type, snp->n_data, signer, SIGN_KEY))){
	    fprintf(stderr, "%s: sign_name(): unable to sign data %d %d\n", 
		    name,type,ret);
	    return ret^0x400;
	  }
	  tput=0;
	}
	if( dp->d_type != T_SIG)  {
	  type = dp->d_type;
	  put_rr( name, dp, ttl);
	  tput++;	  
	}          /* if (dp->d_type  */
      }            /* if( SIGN_BIT_ISSET */
    }              /* for( ndp  */
    if(tput>0 &&(ret=do_signature(name,type, snp->n_data, signer, SIGN_KEY))){
      fprintf(stderr, "%s: sign_name(): unable to sign data+type %d %d\n",
	      name,type,ret);
      return ret^0x200;
    }
    return 0;
  }

/* 
 * make_nullkey( name, dp) creates a null key when there is no key at 
 *  delegation, a warning message is generated. 
 * This routine will return the error codes from db_update(). 
 * 
 */
static int  
make_nullkey(name, mdp, signer)
char *name, *signer;
struct databuf *mdp;
    {
      struct databuf *dp;
      struct namebuf *np;
      static u_char data[] = { 0xc1, 0x00, /* fixed for new 2bit NULL KEY */
				 0xff, 0x1}; 
      int c;
      char *str;

      if( nxt_addname(name, T_KEY, signer) ) {
	return -1;  /* this is when the name is outside the zone
		       are names like a.b.zone and b.zone allowed to be in the
		       same zone ????*/
      }
      fprintf(stderr, "Making NULL KEY for %s\n", name);
      dp = savedata(mdp->d_class, T_KEY, mdp->d_ttl, data, sizeof(data));
      dp->d_zone  = mdp->d_zone;
      dp->d_flags = mdp->d_flags;
      dp->d_cred  = mdp->d_cred;
      dp->d_clev  = dn_count_labels(name);
      dp->d_sec   = mdp->d_sec;
      if ((c = db_update(name, dp, dp, 0x01, hashtab)) != OK) {
	if (c != DATAEXISTS){
	    printf("make_nullkey():update failed %s KEY\n", name);
	}
	free((char*) dp);
	return c;
      }
      dp = nt_lookup( name, T_NXT,  &np);
      str = (char *) dp->d_data +  strlen(dp->d_data) + 1;
      SIGN_BIT_SET( T_KEY, str);
      return  c;
}


static u_int32_t
set_seq_num(tim)
time_t tim;
{
  /* Seq number is set the the current time (seconds since 1969).
   *  People can optionaly supply their own Sequence number
   */ 
  u_int32_t out;
  out = tp.tv_sec;
  printf( "Time %10d -> SQN %9d\n", tim, out);
  return out;
}

static void
set_prefix(ttl, exp_time, signer, footprint,lab )
     int32_t         ttl, exp_time, lab;
     char           *signer;
     int             footprint;
{
	/*
	 * now Here is what I need to do 
	 * 1. create prefix (alg, orig TTL,Expriation TIme, 
	 *                    Time signed, signers name)
         * 2. put binary version in the data structure sig_data.prefix
         *   set the lenght in sig_data.pre_len
	 */
	int             i;
	u_long          opt;
	struct tm      *curr;
	struct timeval  clock;
	char            tmp[200], *pu;

	memset(sig_data.prefix, 0, sizeof(sig_data.prefix));
	/* slot 0 and 1 are for the type */
	sig_data.prefix[SIGN_ALG] = 0x01;	/* put out first field ALG */

	/* now put out the lables */
	sig_data.prefix[SIGN_LABELS] = (u_char) lab;

	pu = (char *) &sig_data.prefix[SIGN_OTTL];
	PUTLONG(ttl, pu);	/* set TTL */
	sig_data.ttl = ttl;
	(void) dprintf(8,(ddt, "Set_prefix: %d ; TTL\n", ttl));

	(void) gettime(&clock);
	if( clock.tv_sec > exp_time) {
	  printf("set_prefix: Expiration time in the past curr= %d expr= %d\n",
		 clock.tv_sec, (int) exp_time);
	  exit(18);
	}

	/* add expiration and signing times to the prefix */
        PUTLONG(exp_time, pu); 
	curr = gmtime((time_t *) &exp_time);
	i = strftime(tmp, sizeof(tmp), 
		     "\t\t%Y%m%d%H%M%S ; Expiration TIME", curr);
	sprintf( tmp, "%d %x	\n",  exp_time, exp_time);

	PUTLONG(clock.tv_sec, pu);
	curr = gmtime((time_t *) &clock.tv_sec);
	i = strftime(tmp, sizeof(tmp), 
		     "\t\t%Y%m%d%H%M%S ; Signing TIME\n\t\t", curr);
	PUTSHORT((u_int16_t) footprint, pu) ;
	/* add uncompressed name of signer in lower case to the prefix */
	opt = _res.options; 
	_res.options |=  RES_TRANS_LOWER;
	i = dn_comp( signer, &sig_data.prefix[SIGN_SIGNER], 
		    sizeof(sig_data.prefix)-SIGN_SIGNER, NULL, NULL);
	_res.options = opt;
	sig_data.pre_len = SIGN_SIGNER + i; 
	sig_data.dat_len = 0; /* make sure there is no old data digested */
}

/* 
 * init_signer( KEY * key)  function to clear the values of context and 
 * set the timer value to something more reasonable than the time
 * of last signature. 
 * This function should only be called at the beginning of signing a zone
 */
void 
init_signer( key) 
KEY * key;
{
  (void) dnssec_sign(SIG_MODE_INIT,                 NULL, 0, key, NULL);
  if( key->axfr) {
    __putshort(T_AXFR, sig_data.prefix); 
    (void) dnssec_sign(SIG_MODE_INIT | SIG_MODE_AXFR, sig_data.prefix, 
		       sig_data.pre_len, key, NULL);
  }
  else 
    (void) dnssec_sign(SIG_MODE_INIT | SIG_MODE_AXFR, NULL, 0 ,key, NULL);
  (void) gettime(&last_time);
  sum_time = 0.0;
  records_signed = 0;
}

/*
 * do_signature(): Routine to generate a signature from the data
 *   stored in the sign_data structure. 
 *   This routine hides the dns data structures but calls a crypto routine
 *   to perform the actual signing of the data 
 *  This routine does do both regular signatures and AXFR signatures, 
 *  on regular signatures if axfr flag is on in key the AXFR of the
 *  data and the SIG RR.
 *  The SIGnatures is added to the data base upon success. 
 */
int 
do_signature(domain, type, sdp, signer, key)
char *domain, *signer;
KEY *key;
short type;
struct databuf *sdp;
{
  struct databuf *dp; 
  static u_char signbuff[128], out_buff[1024];
  double time_to_sign; 
  int len=0, ret=0, dbflags=0x01, siglen;

  if( type == T_SIG) {
    memset(sig_data.data, 0, sig_data.dat_len+8);
    sig_data.dat_len = 0;  /* reset this */ 
    return -5;  /* SIG(SIG) is not allowed */
  }

  __putshort(type, sig_data.prefix);   /* set the type we are doing SIG over */
  sig_data.prefix[SIGN_LABELS] = (u_char) dn_count_labels(domain);

  expand_rrs(domain);
  if( sig_data.dat_len == 0 && type != T_AXFR) 
    return -12; /* nothing to sign */

  /* Do the correct type of signature */
  if( type == T_AXFR && key->axfr) {
    if( (siglen= dnssec_sign( (SIG_MODE_AXFR | SIG_MODE_FINAL), 
		     NULL, 0,key, signbuff)) <= 0) {
      printf ("do_signature():AXFR FINAL Failed %s %s %d\n", domain, signer,
	      siglen);
      return NOT_OK;
    }
  }
  else { 
#ifdef VERIFY_FAIL
    sig_data.data[sig_data.dat_len/2] |= 0x10;
#endif /* VERIFY_FAIL */
    if( (siglen = dnssec_sign( SIG_MODE_INIT,  sig_data.prefix, 
			      sig_data.pre_len, key, NULL)) ||
       ((siglen = dnssec_sign( SIG_MODE_FINAL, sig_data.data, sig_data.dat_len,
			      key, signbuff)) <= 0) ){
      printf ( "do_signature(): Failed generating signature for %s %s %s %d\n",
	      domain, signer,__p_type(type), siglen);
      memset(sig_data.data, 0, sig_data.dat_len+8);
      sig_data.dat_len = dp_rrcount = 0; 
      return NOT_OK; 
    }
    if( verify_flag) { /* test if signature is good */
      if( (ret = dnssec_verify( SIG_MODE_INIT, sig_data.prefix, 
			       sig_data.pre_len, key, signbuff, siglen)) || 
	  (ret = dnssec_verify( SIG_MODE_FINAL, sig_data.data, 
			       sig_data.dat_len, key, signbuff,   siglen))){

	printf("do_signature(): Verify failed %s %s %s ret=%d\n", 
	       domain, signer, __p_type(type), ret);
	memset(sig_data.data, 0, sig_data.dat_len+8);
	sig_data.dat_len = dp_rrcount = 0; 
	return ret; 
      }
    }   /* if verify_flag */
  }
  (void) gettime(&time_now);
  time_to_sign = (double) (time_now.tv_sec - last_time.tv_sec) +
    (double)((time_now.tv_usec-last_time.tv_usec)/1000000.0);
  sum_time += time_to_sign;
  records_signed++;
  last_time  = time_now;
  if( report_success)
    printf("do_signature:Generated success %6.3lf %-5.5s %s\n", time_to_sign,
		 __p_type(type), domain);

  /* Create the SIG RR */
  len = sig_data.pre_len + siglen;
  len = 18;
  memcpy(out_buff, sig_data.prefix, len);
  strncpy( &out_buff[len], signer, strlen(signer));
  len += strlen( signer);
  out_buff[len++] = '.';
  out_buff[len++] = 0x0;
  memcpy(out_buff + len, signbuff, siglen);
  len += siglen;

  /* now store the data in the data base */
  dp = savedata(C_IN, T_SIG, (u_int32_t)sdp->d_ttl, out_buff,  len);
  dp->d_zone  = sdp->d_zone;
  dp->d_flags = 0;   /* sdp->d_flags; this changed in 4.9.5 */
  dp->d_cred  = sdp->d_cred;
  dp->d_clev  = dn_count_labels(domain);
/*  if( domain[0] == '*') dp->d_clev++; */
  if ((ret    = db_update(domain, dp, dp, dbflags, hashtab)) != OK) {
    printf("do_signature: Update Failed %s %d %d\n",
	   domain, sdp->d_type, ret);
    memset(sig_data.data, 0, sig_data.dat_len+8);
    sig_data.dat_len = dp_rrcount = 0; 
    return ret;
  }
  if( type != T_AXFR && key->axfr){ /* Add to AXFR hash */
    append_rr( domain, dp, sig_data.ttl);
    if ((ret = dnssec_sign( (SIG_MODE_AXFR | SIG_MODE_UPDATE), 
	     sig_data.data, sig_data.dat_len, key, NULL)) < 0) {
      printf("do_signature():Failed signing AXFR UPDATE data %s %s %s %d\n", 
	     domain, signer, __p_type( type), ret);
      memset(sig_data.data, 0, sig_data.dat_len+8);
      sig_data.dat_len = dp_rrcount = 0; 
      return ret;
    }
  }
  memset(sig_data.data, 0, sig_data.dat_len+8);
  sig_data.dat_len = dp_rrcount = 0; 
  return OK; 
}
